"""
MIT License

Copyright (c) 2019 YangYun
Copyright (c) 2020 Việt Hùng
Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import List

import cv2
import numpy as np
from tensorflow.keras.utils import Sequence

from .augmentation import mosaic
from ...common import (
    media,
    convert_dataset_to_ground_truth as _convert_dataset_to_ground_truth,
)
from ...common.config import YOLOConfig
from ...common.parser import parse_dataset

_AUGMETATION_CACHE_SIZE = 50


class YOLODataset(Sequence):
    def __init__(
        self,
        config: YOLOConfig,
        dataset_list: str,
        dataset_type: str = "converted_coco",
        image_path_prefix: str = "",
        training: bool = True,
    ):
        self.dataset = parse_dataset(
            dataset_list=dataset_list,
            dataset_type=dataset_type,
            image_path_prefix=image_path_prefix,
        )
        self._metayolos = []
        if config.layer_count["yolo"] > 0:
            for i in range(config.layer_count["yolo"]):
                self._metayolos.append(config.find_metalayer("yolo", i))
        elif config.layer_count["yolo_tpu"] > 0:
            for i in range(config.layer_count["yolo_tpu"]):
                self._metayolos.append(config.find_metalayer("yolo_tpu", i))
        else:
            raise RuntimeError(
                "YOLODataset: model does not have a yolo or yolo_tpu layer"
            )

        self._metanet = config.net
        self._metayolos_np = np.zeros(
            (len(self._metayolos), 7 + len(self._metayolos[-1].mask)),
            dtype=np.float32,
        )
        for i, metayolo in enumerate(self._metayolos):
            self._metayolos_np[i, 0] = metayolo.height
            self._metayolos_np[i, 1] = metayolo.width
            self._metayolos_np[i, 2] = metayolo.channels
            self._metayolos_np[i, 3] = metayolo.classes
            self._metayolos_np[i, 4] = metayolo.label_smooth_eps
            self._metayolos_np[i, 5] = metayolo.max
            self._metayolos_np[i, 6] = metayolo.iou_thresh
            for j, mask in enumerate(metayolo.mask):
                self._metayolos_np[i, 7 + j] = mask

        self._anchors_np = np.zeros(
            len(self._metayolos[-1].anchors) * 2, dtype=np.float32
        )
        for i, anchor in enumerate(self._metayolos[-1].anchors):
            self._anchors_np[2 * i] = anchor[0] / self._metanet.width
            self._anchors_np[2 * i + 1] = anchor[1] / self._metanet.height

        # Data augmentation ####################################################

        self._augmentation: List[str] = []
        if config.net.mosaic:
            self._augmentation.append("mosaic")

        if training and len(self._augmentation) > 0:
            self._augmentation_batch = int(config.net.batch * 0.3)
            self._training = True
        else:
            self._augmentation_batch = 0
            self._training = False

        self._augmentation_cache = [
            self._get_dataset(i) for i in range(_AUGMETATION_CACHE_SIZE)
        ]
        self._augmentation_cache_index = 0

    def _convert_dataset_to_ground_truth(self, dataset_bboxes):
        """
        @param `dataset_bboxes`: [[b_x, b_y, b_w, b_h, class_id], ...]

        @return `groud_truth_one`:
            [Dim(yolo.h, yolo.w, yolo.c + len(mask))] * len(yolo)
        """
        return _convert_dataset_to_ground_truth(
            dataset_bboxes, self._metayolos_np, self._anchors_np
        )

    def _convert_dataset_to_image_and_bboxes(self, dataset):
        """
        @param dataset: [image_path, [[x, y, w, h, class_id], ...]]

        @return image, bboxes
            image: 0.0 ~ 1.0, Dim(1, height, width, channels)
        """
        # pylint: disable=bare-except
        try:
            image = cv2.imread(dataset[0])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        except:
            return None, None

        resized_image, resized_bboxes = media.resize_image(
            image,
            target_shape=self._metanet.input_shape,
            ground_truth=dataset[1],
        )
        resized_image = np.expand_dims(resized_image / 255.0, axis=0)

        return resized_image, resized_bboxes

    def _get_dataset(self, index: int):
        offset = 0
        for offset in range(5):
            image, bboxes = self._convert_dataset_to_image_and_bboxes(
                self.dataset[(index + offset) % len(self.dataset)]
            )
            if image is None:
                offset += 1
            else:
                return image, bboxes

        raise FileNotFoundError("Failed to find images")

    def __getitem__(self, index):
        """
        @return
            `images`: Dim(batch, height, width, channels)
            `groud_truth_one`:
                [Dim(batch, yolo.h, yolo.w, yolo.c + len(mask))] * len(yolo)
        """

        batch_x = []
        # [[gt_one, gt_one, ...],
        #  [gt_one, gt_one, ...], ...]
        batch_y = [[] for _ in range(len(self._metayolos))]

        start_index = index * self._metanet.batch

        for i in range(self._metanet.batch - self._augmentation_batch):
            image, bboxes = self._get_dataset(start_index + i)
            self._augmentation_cache[self._augmentation_cache_index] = (
                image,
                bboxes,
            )
            self._augmentation_cache_index = (
                self._augmentation_cache_index + 1
            ) % _AUGMETATION_CACHE_SIZE

            batch_x.append(image)
            ground_truth = self._convert_dataset_to_ground_truth(bboxes)
            for j in range(len(self._metayolos)):
                batch_y[j].append(ground_truth[j])

        for i in range(self._augmentation_batch):
            augmentation = self._augmentation[
                np.random.randint(0, len(self._augmentation))
            ]

            image = None
            bboxes = None
            if augmentation == "mosaic":
                image, bboxes = mosaic(
                    *[
                        self._augmentation_cache[
                            np.random.randint(
                                0,
                                _AUGMETATION_CACHE_SIZE,
                            )
                        ]
                        for _ in range(4)
                    ]
                )

            batch_x.append(image)
            ground_truth = self._convert_dataset_to_ground_truth(bboxes)
            for j in range(len(self._metayolos)):
                batch_y[j].append(ground_truth[j])

        return np.concatenate(batch_x, axis=0), [
            np.stack(y, axis=0) for y in batch_y
        ]

    def __len__(self):
        return len(self.dataset) // (
            self._metanet.batch - self._augmentation_batch
        )



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Any

from .base_layer import BaseLayer


class ConvolutionalLayer(BaseLayer):
    def __init__(self, index: int, type_index: int):
        super().__init__(
            index=index, type_index=type_index, type_name="convolutional"
        )
        self._activation = "logistic"
        self._batch_normalize = 0
        self._filters = 1
        self._pad = False
        self._padding = 0
        self._size = 1
        self._stride = 1

    @property
    def activation(self) -> str:
        return self._activation

    @property
    def batch_normalize(self) -> int:
        return self._batch_normalize

    @property
    def bflops(self) -> float:
        """
        kernel: size x size x input_filter x output_filter
            >>
                bias: output_filter
                or
                batch_normalization: output_filter x 4
        """
        return (
            2
            * self._size
            * self._size
            * self._input_shape[-1]
            * self._output_shape[-1]
            * self._output_shape[0]
            * self._output_shape[1]
        ) / 1e9

    @property
    def filters(self) -> int:
        return self._filters

    @property
    def pad(self) -> bool:
        return self._pad

    @property
    def padding(self) -> int:
        if self._pad:
            return self._size // 2
        return self._padding

    @property
    def size(self) -> int:
        return self._size

    @property
    def stride(self) -> int:
        return self._stride

    def __repr__(self) -> str:
        rep = f"{self.index:4}  "
        rep += f"{self.type[:5]}_"
        rep += f"{self.type_index:<3}  "
        rep += f"{self.filters:4}     "
        rep += f"{self.size:2} x{self.size:2} /{self.stride:2}     "
        rep += f"{self.input_shape[0]:4} "
        rep += f"x{self.input_shape[1]:4} "
        rep += f"x{self.input_shape[2]:4} -> "
        rep += f"{self.output_shape[0]:4} "
        rep += f"x{self.output_shape[1]:4} "
        rep += f"x{self.output_shape[2]:4}  "
        rep += f"{self.bflops:6.3f}"
        return rep

    def __setitem__(self, key: str, value: Any):
        if key in ("activation",):
            self.__setattr__(f"_{key}", str(value))
        elif key in ("batch_normalize", "filters", "padding", "size", "stride"):
            self.__setattr__(f"_{key}", int(value))
        elif key in ("pad",):
            self.__setattr__(f"_{key}", bool(int(value)))
        elif key == "input_shape":
            self.__setattr__(f"_{key}", value)
            self._output_shape = (
                self._input_shape[0] // self._stride,
                self._input_shape[1] // self._stride,
                self._filters,
            )
        else:
            raise KeyError(f"'{key}' is not supported")



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from os import makedirs, path
import shutil

import cv2
import numpy as np


def create_mAP_input_files(
    yolo,
    dataset,
    mAP_path: str,
    images_optional: bool = False,
    num_sample: int = None,
):
    """
    Ref: https://github.com/Cartucho/mAP
    gt: name left top right bottom
    dr: name confidence left top right bottom

    @param `yolo`
    @param `dataset`
    @param `mAP_path`
    @param `images_optional`: If `True`, images are copied to the
            `mAP_path`.
    @param `num_sample`: Number of images for mAP. If `None`, all images in
            `data_set` are used.
    """
    input_path = path.join(mAP_path, "input")

    if path.exists(input_path):
        shutil.rmtree(input_path)
    makedirs(input_path)

    gt_dir_path = path.join(input_path, "ground-truth")
    dr_dir_path = path.join(input_path, "detection-results")
    makedirs(gt_dir_path)
    makedirs(dr_dir_path)

    img_dir_path = ""
    if images_optional:
        img_dir_path = path.join(input_path, "images-optional")
        makedirs(img_dir_path)

    max_dataset_size = len(dataset.dataset)

    if num_sample is None:
        num_sample = max_dataset_size

    if num_sample > max_dataset_size:
        num_sample = max_dataset_size

    for i in range(num_sample):
        # image_path, [[x, y, w, h, class_id], ...]
        image_path, gt_bboxes = dataset.dataset[i % max_dataset_size].copy()

        if images_optional:
            target_path = path.join(img_dir_path, "image_{}.jpg".format(i))
            shutil.copy(image_path, target_path)

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        height, width, _ = image.shape

        gt_bboxes = gt_bboxes * np.array([width, height, width, height, 1])

        # ground-truth
        with open(
            path.join(gt_dir_path, "image_{}.txt".format(i)),
            "w",
        ) as fd:
            for xywhc in gt_bboxes:
                # name left top right bottom
                class_name = yolo.config.names[int(xywhc[4])].replace(" ", "_")
                left = int(xywhc[0] - xywhc[2] / 2)
                top = int(xywhc[1] - xywhc[3] / 2)
                right = int(xywhc[0] + xywhc[2] / 2)
                bottom = int(xywhc[1] + xywhc[3] / 2)
                fd.write(
                    "{} {} {} {} {}\n".format(
                        class_name, left, top, right, bottom
                    )
                )

        # predict
        pred_bboxes = yolo.predict(image)
        # Dim(-1, (x,y,w,h,o, cls_id0, prob0, cls_id1, prob1))
        pred_bboxes = pred_bboxes * np.array(
            [width, height, width, height, 1, 1, 1, 1, 1]
        )

        # detection-results
        with open(
            path.join(dr_dir_path, "image_{}.txt".format(i)),
            "w",
        ) as fd:
            for xywhcp in pred_bboxes:
                # name confidence left top right bottom
                class_name = yolo.config.names[int(xywhcp[5])].replace(" ", "_")
                probability = xywhcp[6]
                if probability < 0.01:
                    continue
                left = int(xywhcp[0] - xywhcp[2] / 2)
                top = int(xywhcp[1] - xywhcp[3] / 2)
                right = int(xywhcp[0] + xywhcp[2] / 2)
                bottom = int(xywhcp[1] + xywhcp[3] / 2)
                fd.write(
                    "{} {} {} {} {} {}\n".format(
                        class_name, probability, left, top, right, bottom
                    )
                )



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from os import path
import time

import cv2
import numpy as np

from . import media
from .config import YOLOConfig
from ._common import (
    fit_to_original as _fit_to_original,
    yolo_diou_nms as _yolo_diou_nms,
)


class BaseClass:
    def __init__(self):
        self.config = YOLOConfig()

    @staticmethod
    def yolo_diou_nms(candidates: np.ndarray, beta_nms: float) -> np.ndarray:
        """
        Warning!
            - change order
            - change c0 -> p(c0)

        @param `candidates`: Dim(-1, 5 + len(classes))
        @param `beta_nms`: rdiou = pow(d / c, beta1);

        @return `pred_bboxes`
            Dim(-1, (x,y,w,h,o, cls_id0, prob0, cls_id1, prob1))
        """
        return _yolo_diou_nms(candidates=candidates, beta1=beta_nms)

    def fit_to_original(
        self, pred_bboxes: np.ndarray, origin_height: int, origin_width: int
    ):
        """
        Warning! change pred_bboxes directly

        @param `pred_bboxes`
            Dim(-1, (x,y,w,h,o, cls_id0, prob0, cls_id1, prob1))
        """
        _fit_to_original(
            pred_bboxes,
            self.config.net.height,
            self.config.net.width,
            origin_height,
            origin_width,
        )

    def resize_image(self, image, ground_truth=None):
        """
        @param image:        Dim(height, width, channels)
        @param ground_truth: [[center_x, center_y, w, h, class_id], ...]

        @return resized_image or (resized_image, resized_ground_truth)

        Usage:
            image = yolo.resize_image(image)
            image, ground_truth = yolo.resize_image(image, ground_truth)
        """
        input_shape = self.config.net.input_shape
        return media.resize_image(
            image,
            target_shape=input_shape,
            ground_truth=ground_truth,
        )

    def draw_bboxes(self, image, pred_bboxes):
        """
        @parma `image`:  Dim(height, width, channel)
        @parma `pred_bboxes`
            Dim(-1, (x,y,w,h,o, cls_id0, prob0, cls_id1, prob1))

        @return drawn_image

        Usage:
            image = yolo.draw_bboxes(image, bboxes)
        """
        return media.draw_bboxes(image, pred_bboxes, names=self.config.names)

    #############
    # Inference #
    #############

    def predict(self, frame: np.ndarray):
        # pylint: disable=unused-argument, no-self-use
        return [[0.0, 0.0, 0.0, 0.0, -1]]

    def inference(
        self,
        media_path,
        is_image: bool = True,
        cv_apiPreference=None,
        cv_frame_size: tuple = None,
        cv_fourcc: str = None,
        cv_waitKey_delay: int = 1,
    ):
        if isinstance(media_path, str) and not path.exists(media_path):
            raise FileNotFoundError("{} does not exist".format(media_path))

        cv2.namedWindow("result", cv2.WINDOW_AUTOSIZE)

        if is_image:
            frame = cv2.imread(media_path)
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

            start_time = time.time()
            bboxes = self.predict(frame)
            exec_time = time.time() - start_time
            print("time: {:.2f} ms".format(exec_time * 1000))

            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
            image = self.draw_bboxes(frame, bboxes)
            cv2.imshow("result", image)
        else:
            if cv_apiPreference is None:
                cap = cv2.VideoCapture(media_path)
            else:
                cap = cv2.VideoCapture(media_path, cv_apiPreference)

            if cv_frame_size is not None:
                cap.set(cv2.CAP_PROP_FRAME_WIDTH, cv_frame_size[0])
                cap.set(cv2.CAP_PROP_FRAME_HEIGHT, cv_frame_size[1])

            if cv_fourcc is not None:
                cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*cv_fourcc))

            prev_time = time.time()
            if cap.isOpened():
                while True:
                    try:
                        is_success, frame = cap.read()
                    except cv2.error:
                        continue

                    if not is_success:
                        break

                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

                    predict_start_time = time.time()
                    bboxes = self.predict(frame)
                    predict_exec_time = time.time() - predict_start_time

                    frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
                    image = self.draw_bboxes(frame, bboxes)
                    curr_time = time.time()

                    cv2.putText(
                        image,
                        "preidct: {:.2f} ms, fps: {:.2f}".format(
                            predict_exec_time * 1000,
                            1 / (curr_time - prev_time),
                        ),
                        org=(5, 20),
                        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=0.6,
                        color=(50, 255, 0),
                        thickness=2,
                        lineType=cv2.LINE_AA,
                    )
                    prev_time = curr_time

                    cv2.imshow("result", image)
                    if cv2.waitKey(cv_waitKey_delay) & 0xFF == ord("q"):
                        break

            cap.release()

        print("YOLOv4: Inference is finished")
        while cv2.waitKey(10) & 0xFF != ord("q"):
            pass
        cv2.destroyWindow("result")
